#ifndef __Zr_H__
#define __Zr_H__

#include <iostream>
#include "G.h"
#include <gmp.h>
#undef __cplusplus
extern "C" {
#include <relic/relic.h>
}
#define __cplusplus 1

using namespace std;

class Zr {
public:
  Zr() { init(); }
  ~Zr() { bn_free(g); bn_free(ord);  }

  //Basic initialization to zero
  Zr(const Pairing &e);
  
  //random = true  -> random value
  //random = false -> zero
  Zr(const Pairing &e, bool random) { initMaybeRandom(random); }
  Zr(bool random) { initMaybeRandom(random); }

  //Copy constructor
  Zr(const Zr& other);

  //Create an element from a long int
  Zr(const Pairing &e, long long int i) { initWithNum((unsigned long long int) i); }
  Zr(const Pairing &e, long int i) { initWithNum((unsigned long long int) i); }
  Zr(const Pairing &e, int i) { initWithNum((unsigned long long int) i); }
  Zr(long long int i) { initWithNum((unsigned long long int) i); }
  Zr(long int i) { initWithNum((unsigned long long int) i); }
  Zr(int i) { initWithNum((unsigned long long int) i); }

  //From hash
  Zr(const Pairing &e, const void *data, unsigned short len);

  //Init from string
  Zr(string s, unsigned short base = 10);

  //PBC-style import
  Zr(const Pairing &e, const unsigned char *data,
         unsigned short len, unsigned short base = 16);

  //Initialize from another, but with a different value (for compatibility)
  Zr(const Zr& other, long int i) { initWithNum((unsigned long long int) i); }

  //Arithmetic operators
  Zr& operator+=(const Zr& other);
  Zr& operator*=(const Zr& other);
  Zr& operator-=(const Zr& other); 
  Zr& operator/=(const Zr& other);
  Zr& operator^=(const Zr& other);

  const Zr operator+(const Zr &other) const { return Zr(*this) += other; }
  const Zr operator*(const Zr &other) const { return Zr(*this) *= other; }
  const Zr operator-(const Zr &other) const { return Zr(*this) -= other; }
  const Zr operator/(const Zr &other) const { return Zr(*this) /= other; }
  const Zr operator^(const Zr &other) const { return Zr(*this) ^= other; }

  bool operator==(const Zr &other);

  const Zr square() {
    Zr a = Zr();
    a.setElement(g);
    return a * a;
  }
  const Zr inverse(bool additive=false) const {
    Zr tempz;
    if (additive) {
      tempz = Zr(0);
      tempz -= *this;
    }
    else {
      tempz = Zr(1);
      tempz /= *this;
    }
    return tempz;
  }
  void operator=(const Zr &other);

  //String and I/O methods
  string toString (unsigned short base=10) const;
  string decimalToString () const { return toString(10); }

  void dump(FILE *f, const char *label="",
			 unsigned short base=10) const {
    if (base < 256) {
      if (label) fprintf(f, "%s: ", label);
      fprintf(f,"%s\n",toString(base).c_str());
    }
    else {
      char o[1000]; toBin(o);
      for (int i = 0; i < 32; i++) { fputc(o[i],f); }
    }
  }

  void fromBin(const char* input);
  void toBin(char* output) const;

  void fromString(string input, unsigned short base = 10) {
    istringstream iss(input); 
    fromStream(iss,base);
  }

  friend inline ostream& operator<<(ostream &out, const Zr& x) {
    out << x.toString(10);
    return out;
  }

  friend inline istream& operator>>(istream &in, Zr& out) {
    out.fromStream(in,10);
    return in;
  }

  void fromStream(istream &in, int base);

  //Helper functions
  void getElement (bn_t out) const {
    bn_copy(out,const_cast<bn_st*>(g));
  }
  void setElement ( bn_t in) {
    bn_copy(g,in);
  }
  bool isIdentity(bool additive) { return *this == Zr(additive ? 0 : 1); }
  bool isElementPresent() const { return true; }
  Zr randomize();
  void init();

private:
  bn_t g;
  bn_t ord;
  void initWithNum(unsigned long long int i);
  void initMaybeRandom(bool random);
};


#endif
